/*  This file should be edited with 4-column tabs! */
/*  Author: Mark Moraes <moraes@csri.toronto.edu> */

/*LINTLIBRARY*/

#include "defs.h"
#include "globals.h"

RCSID("$Header: /j/moraes/shmem/malloc/malloc/RCS/valloc.c,v 1.5 89/10/31 02:02:16 moraes Exp $");

/* Just following the Sun manual page here */
univptr_t
valloc(size)
size_t size;
{
	static size_t pagesz = 0;

	if (pagesz == 0)
		pagesz = (size_t) getpagesize();
	return(memalign(pagesz, size));
}
