#ifdef __STDC__
#include <stddef.h>
/*#include <stdlib.h>*/
#define u_int size_t
#else
#define u_int unsigned int
#endif

/* For now, both ansi and normal C need this - one day... */
extern int fputs();
extern char *memset();
extern int fprintf();
extern int atoi();
extern void exit();
extern int write();

#include <stdio.h>
#include "malloc.h"

int
main(argc, argv)
char **argv;
int argc;
{
	char *cp;
	int nbytes;

	if (argc != 2) {
		(void) fprintf(stderr, "Usage: %s nbytes", argv[0]);
		exit(1);
	}

	nbytes = atoi(argv[1]);
	cp = (char *) malloc(nbytes);
	cp[nbytes] = 'a';
	mal_verify(1);
	/* We aren't going to get here, y'know... */
	return 0;
}
