/*  This file should be edited with 4-column tabs! */
/*  Author: Mark Moraes <moraes@csri.toronto.edu> */

/*LINTLIBRARY*/

#include "defs.h"
#include "globals.h"

RCSID("$Header: /j/moraes/shmem/malloc/malloc/RCS/setstats.c,v 1.6 89/10/31 02:02:02 moraes Exp $");

/*
 *  Sets the file used for verbose statistics to 'fd'. Does no
 *  verification whatsoever on the file descriptor
 */
void
mal_setstatsfile(fd)
FILE * fd;
{
	_malloc_statsfile = fd;
	/*
	 *  This file descriptor had better not have been written to before
	 *  this
	 */
	(void) setvbuf(fd, (char *) 0, _IONBF, 0);
}
