/*
 * getdtablesize -- call getrlimit
 */

#include "hostenv.h"
#ifndef	USE_GETDTABLESIZE

#include <sys/resource.h>
getdtablesize()
{
        struct rlimit res;
        int stat;
        stat = getrlimit(RLIMIT_NOFILE, &res);
        if (stat < 0) return(-1);
        return(res.rlim_cur);
}

#endif	/* !USE_GETDTABLESIZE */
