/*
 *	Copyright 1988 by Rayan S. Zachariassen, all rights reserved.
 *	This will be free software, but only when it is finished.
 */

/*
 * Support routine for the mail submission interface.  This file gets linked
 * in if the application doesn't define its own mail_alloc() routine.
 */

char *
mail_alloc(nbytes)
	unsigned int nbytes;
{
	extern char *malloc();

	return malloc(nbytes);
}

int
mail_free(s)
	char *s;
{
	extern int free();

	return free(s);
}
