/*
 *	Copyright 1991 by Rayan S. Zachariassen, all rights reserved.
 *	This will be free software, but only when it is finished.
 */

#include "hostenv.h"
#ifndef	USE_STRERROR
char *
strerror(num)
	int num;
{
	extern char *sys_errlist[];

	return sys_errlist[num];
}
#endif	/* !USE_STRERROR */
